package server

import (
	"context"
	"errors"
	odysseus "github.com/odysseus/odysseus-protocol/gen/proto/go/base/v1"
	omanager "github.com/odysseus/odysseus-protocol/gen/proto/go/nodemanager/v2"
	"github.com/odysseus/scheduler/config"
	"github.com/odysseus/scheduler/types"
	log "github.com/sirupsen/logrus"
	"google.golang.org/grpc"
	"google.golang.org/grpc/credentials/insecure"
	"strconv"
	"strings"
	"time"
)

var (
	ErrDispatchFailed = errors.New("dispatch to nodemanager failed")
)

func newManagerClient(endpoint string) (omanager.NodeManagerServiceClient, error) {
	client, err := grpc.Dial(endpoint,
		grpc.WithTransportCredentials(insecure.NewCredentials()),
		grpc.WithDefaultCallOptions(
			grpc.MaxCallRecvMsgSize(1024*1024*1024),
			grpc.MaxCallSendMsgSize(1024*1024*1024)),
	)
	if err != nil {
		return nil, err
	}
	manager := omanager.NewNodeManagerServiceClient(client)
	return manager, nil

}

func parseWorkerNmValue(nmValue string) (string, int64) {
	split := "#"
	strs := strings.Split(nmValue, split)
	if len(strs) == 2 {
		endpoint := strs[0]
		timestamp, _ := strconv.ParseInt(strs[1], 10, 64)
		return endpoint, timestamp
	}
	return "", 0
}

func (n *Node) DispatchTask(ctx context.Context, worker types.Worker, task *odysseus.TaskContent) error {
	l := log.WithField("task-id", task.TaskId)
	l.WithFields(log.Fields{
		"worker":      worker.Workerid,
		"managerList": worker.Managers,
	}).Debug("dispatch task to worker")
	var shouldAddBack = false
	defer func(w types.Worker) {
		if shouldAddBack && n.poper.CanAddBack() {
			n.poper.AddBack(w)
		}
	}(worker)

	for _, manager := range worker.Managers {
		endpoint, updateTime := parseWorkerNmValue(manager)
		if time.Now().Unix()-updateTime > int64(config.GetConfig().MaxNmUpdateEx) {
			l.WithField("manager", manager).Warn("ignore the manager because of not update tm")
			continue
		}

		client, err := newManagerClient(endpoint)
		if err != nil {
			l.WithFields(log.Fields{
				"manager": endpoint,
				"error":   err,
			}).Error("connect to manager failed")
			continue
		}
		_, err = client.DispatchTask(ctx, &omanager.DispatchTaskRequest{
			Miner:    worker.Workerid,
			TaskData: task,
		})
		if err != nil {
			l.WithFields(log.Fields{
				"manager": endpoint,
				"error":   err,
			}).Error("dispatch to manager failed")
			if strings.HasSuffix(err.Error(), "deadline exceeded") {
				shouldAddBack = true
				return types.ErrTimeout
			}
			continue
		}
		return nil

	}
	return ErrDispatchFailed
}
